# -*- coding: utf-8 -*-
#
# Message tagging service is an event-driven service to tag build.
# Copyright (C) 2019  Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Chenxiong Qi <cqi@redhat.com>, Valerij Maljulin <vmaljuli@redhat.com>


import importlib.machinery
import os
import sys


running_tests = any('py.test' in arg for arg in sys.argv)


class Config:
    """
    This class allows to instanciate configuration. It has 3 types of properties
        (starting from the least priority):
    - default properties - set as members of dictionary __defaults to this class
        (common for all instances)
    - class properties - properties of configuration class from configuration file (__conf_class)
    - overrided properties - properties changed in runtime
    All properties could be read using instance.name or instance['name'] form.
    Setting a new value to instance['name'] will override the value of property
        (but only for this particular instance)
    """
    _defaults = {}

    def __init__(self, profile=None, config_file=None, config_class=None):
        """
        Initialize config, read sets configuration class from configuration file
        :param profile: to be set to config file name
        :type profile: str
        :param config_file: force configuration file path (overrides profile setting if present)
        :type config_file: str
        :param config_class: force class name in configuration file
        :type config_class: str
        """
        self._conf_class = None  # set lazily later
        self._overrides = {}
        self.load_new_config(profile, config_file, config_class)

    @staticmethod
    def get_config_file(profile):
        """
        Trying to get information about config file. Lookup through:
        1. Value of MTS_CONFIG_FILE environmental variable
        2. Try development configuration if MTS_DEV is set or in tests mode
        3. Default configuration path /etc/mts/config.py
        :param profile: profile name to be a part of config file
        :type profile: str
        :return: file path
        :rtype: str
        """
        config_file = os.environ.get('MTS_CONFIG_FILE')
        if config_file:
            return config_file
        elif 'MTS_DEV' in os.environ or running_tests:
            # Use {project root directory}/conf/config.py
            return os.path.realpath(
                os.path.join(os.path.dirname(__file__), '..', 'conf', 'config.py'))
        else:
            if profile:
                return f'/etc/mts/config.{profile}.py'
            else:
                return '/etc/mts/config.py'

    @staticmethod
    def get_config_class_name():
        """
        Determine a class name according to running mode
        :return: class name
        :rtype: str
        """
        if running_tests:
            return 'TestConfiguration'
        elif 'MTS_DEV' in os.environ:
            return 'DevConfiguration'
        else:
            return 'BaseConfiguration'

    def update(self, new_val_dict=None, **kwargs):
        """
        Support update configuration with the dictionary or keyword arguments
        :param new_val_dict: dictionary containing new configuration values
        :type new_val_dict: dict
        :param kwargs: keyword-styled new configuration values
        :return:
        """
        if new_val_dict and isinstance(new_val_dict, dict):
            self._overrides.update(new_val_dict)
        self._overrides.update(kwargs)

    def reset(self):
        """
        Reset overrides
        :return:
        """
        self._overrides.clear()

    def load_new_config(self, profile=None, config_file=None, config_class=None,
                        keep_overrides=False):
        """
        Reloads a base file from a config class
        :param profile: to be set to config file name
        :type profile: str
        :param config_file: force configuration file path (overrides profile setting if present)
        :type config_file: str
        :param config_class: force class name in configuration file
        :type config_class: str
        :param keep_overrides: keep overrided values if True
        :type keep_overrides: bool
        :return:
        """
        if not config_file:
            config_file = self.get_config_file(profile=profile)
        loader = importlib.machinery.SourceFileLoader('mts_conf', config_file)
        mod = loader.load_module()
        if not config_class:
            config_class = self.get_config_class_name()
        if getattr(mod, config_class, None) is not None:
            self._conf_class = getattr(mod, config_class)
        else:
            raise AttributeError(f'Configuration class {config_class} '
                                 f'not found in configuration file {config_file}')
        if not keep_overrides:
            self._overrides.clear()

    def __getattr__(self, item):
        """
        Allows to access configuration parameters as instance.item
        :param item: item name
        :type item: str
        :return: item value
        """
        return self[item]

    def __getitem__(self, item):
        """
        Allows to access configuration parameters as instance['item']
        :param item: item name
        :return: item value
        """
        # check overrides first:
        if item in self._overrides:
            return self._overrides[item]

        # trying configuration class then
        try:
            return getattr(self._conf_class, item)
        except AttributeError:
            pass

        # fallback to defaults if any
        if item in self._defaults:
            return self._defaults[item]

        raise KeyError(f'{self._conf_class.__name__} has no config {item}.')

    def __setitem__(self, key, value):
        """
        Value can be overrided using instance['key'] = value
        :param key: item name
        :type key: str
        :param value: new value
        :return: None
        """
        self._overrides[key] = value

    @property
    def conf_class(self):
        return self._conf_class



# -*- coding: utf-8 -*-
#
# Message tagging service is an event-driven service to tag build.
# Copyright (C) 2019  Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Chenxiong Qi <cqi@redhat.com>

import logging
import json

from message_tagging_service import conf, monitor

logger = logging.getLogger(__name__)


def publish(topic, msg):
    """
    Publish a single message to a given backend, and return

    :param str topic: the topic of the message (e.g. module.state.change)
    :param dict msg: the message contents of the message (typically JSON)
    :return: the value returned from underlying backend "send" method.
    """
    backend = conf.messaging_backend
    try:
        handler = _messaging_backends[backend]['publish']
    except KeyError:
        raise KeyError(f'No messaging backend found for {backend}')
    try:
        return handler(topic, msg)
    except Exception:
        monitor.messages_notify_errors_counter.inc()
        logger.exception('Failed to send message to topic %s: %s', topic, msg)


def _fedmsg_publish(topic, msg):
    # fedmsg doesn't really need access to conf, however other backends do
    import fedmsg
    config = conf.messaging_backends['fedmsg']
    if conf.dry_run:
        logger.info("DRY-RUN: fedmsg.publish('%s', msg=%s, modname='%s')",
                    topic, msg, config['service'])
    else:
        return fedmsg.publish(topic, msg=msg, modname=config['service'])


def _rhmsg_publish(topic, msg):
    """Send message to Unified Message Bus

    :param str topic: the topic where message will be sent to, e.g.
        ``build.tagged``.
    :param dict msg: the message that will be sent.
    """
    import proton
    from rhmsg.activemq.producer import AMQProducer

    config = conf.messaging_backends['rhmsg']
    producer_config = {
        'urls': config['brokers'],
        'certificate': config['certificate'],
        'private_key': config['private_key'],
        'trusted_certificates': config['ca_cert'],
    }
    with AMQProducer(**producer_config) as producer:
        prefix = config['topic_prefix'].rstrip('.')
        topic = f'{prefix}.{topic}'
        producer.through_topic(topic)

        outgoing_msg = proton.Message()
        outgoing_msg.body = json.dumps(msg)
        if conf.dry_run:
            logger.info('DRY-RUN: AMQProducer.send(%s) through topic %s',
                        outgoing_msg, topic)
        else:
            producer.send(outgoing_msg)


_messaging_backends = {
    'fedmsg': {
        'publish': _fedmsg_publish
    },
    'rhmsg': {
        'publish': _rhmsg_publish
    }
}



# -*- coding: utf-8 -*-
#
# Message tagging service is an event-driven service to tag build.
# Copyright (C) 2019  Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Chenxiong Qi <cqi@redhat.com>

import os

from mock import patch
from message_tagging_service import config

test_config = os.path.join(os.path.dirname(__file__),
                           'data',
                           'config.py')


class TestConfig(object):

    @patch.dict('os.environ', values={'MTS_CONFIG_FILE': test_config})
    def test_use_specified_config_file(self):
        conf = config.Config()
        assert conf.test

    @patch.dict('os.environ', values={'MTS_DEV': '1'})
    def test_test_config(self):
        conf = config.Config()
        assert 'TestConfiguration' == conf.conf_class.__name__

    @patch.dict('os.environ', values={'MTS_DEV': '1'})
    @patch.object(config, 'running_tests', new=False)
    def test_dev_config(self):
        conf = config.Config()
        assert 'DevConfiguration' == conf.conf_class.__name__

    @patch.dict('os.environ', clear=True)
    @patch.object(config, 'running_tests', new=False)
    @patch('importlib.machinery')
    def test_use_installed_config(self, machinery):
        conf = config.Config()

        machinery.SourceFileLoader.assert_called_once_with(
            'mts_conf',
            '/etc/mts/config.py')

        loader = machinery.SourceFileLoader.return_value
        mod = loader.load_module.return_value
        assert conf.conf_class == mod.BaseConfiguration

    @patch.dict('os.environ', clear=True)
    @patch.object(config, 'running_tests', new=False)
    @patch('importlib.machinery')
    def test_config_profile(self, machinery):
        profile_name = 'abc123'
        conf = config.Config(profile=profile_name)

        machinery.SourceFileLoader.assert_called_once_with(
            'mts_conf',
            f'/etc/mts/config.{profile_name}.py')

        loader = machinery.SourceFileLoader.return_value
        mod = loader.load_module.return_value
        assert conf.conf_class == mod.BaseConfiguration

    @patch.dict('os.environ', values={'MTS_CONFIG_FILE': test_config})
    def test_overriding(self):
        new_value = 'dddeeefff123'
        test_val2 = 3
        test_val3 = 10
        conf = config.Config()
        import tests.data.config as conf_data
        assert conf.test_val1 == conf_data.TestConfiguration.test_val1
        assert conf['test_val1'] == conf_data.TestConfiguration.test_val1
        conf['test_val1'] = new_value
        assert conf.test_val1 == new_value
        assert conf['test_val1'] == new_value
        conf.update({'test_val2': test_val2})
        conf.update(test_val3=test_val3)
        assert conf.test_val2 == test_val2
        assert conf.test_val3 == test_val3
        conf.reset()
        assert conf.test_val1 == conf_data.TestConfiguration.test_val1



# -*- coding: utf-8 -*-
#
# Message tagging service is an event-driven service to tag build.
# Copyright (C) 2019  Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Chenxiong Qi <cqi@redhat.com>

import pytest

from mock import patch, Mock
from message_tagging_service import utils
from requests.exceptions import HTTPError


class TestRetrieveModulemdContent(object):
    """Test utils.retrieve_modulemd_content"""

    @patch.object(utils.conf, 'mbs_api_url', new='https://mbs.local/')
    @patch('requests.get')
    def test_retrieve_the_content(self, get):
        get.return_value = Mock(status_code=200)
        fake_modulemd = 'modulemd conent'
        get.return_value.json.return_value = {
            'modulemd': fake_modulemd,
        }

        modulemd = utils.retrieve_modulemd_content(1)

        assert fake_modulemd == modulemd
        get.assert_called_once_with(
            'https://mbs.local/module-builds/1',
            params={'verbose': True})

    @patch.object(utils.conf, 'mbs_api_url', new='https://mbs.local/')
    @patch('requests.get')
    def test_raise_error_if_failed_to_get_module(self, get):
        get.return_value.raise_for_status.side_effect = HTTPError('error')
        pytest.raises(HTTPError, utils.retrieve_modulemd_content, 1)



# -*- coding: utf-8 -*-
#
# Message tagging service is an event-driven service to tag build.
# Copyright (C) 2019  Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Chenxiong Qi <cqi@redhat.com>

from flask import Flask, Response
from prometheus_client import CONTENT_TYPE_LATEST

from message_tagging_service.monitor import generate_metrics_report

app = Flask(__name__)


@app.route('/', methods=['GET'])
def index():
    return Response('Welcome to Message-Tagging-Service (aka MTS)')


@app.route('/monitor/metrics', methods=['GET'])
def metrics():
    return Response(generate_metrics_report(),
                    content_type=CONTENT_TYPE_LATEST)



# Setup fedmsg logging.
# See the following for constraints on this format https://bit.ly/Xn1WDn
bare_format = "[%(asctime)s][%(name)10s %(levelname)7s] %(message)s"

config = {
    'logging': {
        'version': 1,
        'formatters': {
            'bare': {
                'datefmt': '%Y-%m-%d %H:%M:%S',
                'format': bare_format
            },
        },
        'handlers': {
            'console': {
                "class": "logging.StreamHandler",
                "formatter": "bare",
                "level": "DEBUG",
                "stream": "ext://sys.stdout",
            },
        },
        'loggers': {
            'fedmsg': {
                "level": "DEBUG",
                "propagate": False,
                "handlers": ["console"],
            },
            'moksha': {
                "level": "DEBUG",
                "propagate": False,
                "handlers": ["console"],
            },
        },
    },
}



# -*- coding: utf-8 -*-

import os

config = {
    "validate_signatures": False,
    "active": True,
    "environment": "prod",
    "endpoints": {
        "fedora-infrastructure": [
            # Just listen to staging for now, not to production (spam!)
            # "tcp://hub.fedoraproject.org:9940",
            "tcp://stg.fedoraproject.org:9940",
        ],
        'relay_outbound': ["tcp://127.0.0.1:4001"],
    },
    'relay_inbound': ["tcp://127.0.0.1:2003"],

    'mts-consumer': True,
}

if 'MTS_USE_STOMP' in os.environ:
    config.update({
        'zmq_enabled': False,
        'stomp_heartbeat': 1000,
        # UMB brokers. Could be a comma-separated string of broker URIs.
        'stomp_uri': os.environ['MTS_STOMP_URI'],
        # File name with absolute path to certificate file.
        'stomp_ssl_crt': os.environ['MTS_STOMP_SSL_CRT'],
        # File name with absolute path to private key file.
        'stomp_ssl_key': os.environ['MTS_STOMP_SSL_KEY'],
    })



# -*- coding: utf-8 -*-

import os


class BaseConfiguration:
    dry_run = os.environ.get('MTS_DRY_RUN', False)
    mbs_api_url = 'https://mbs.fedoraproject.org/module-build-service/1/'

    koji_profile = 'koji'

    # User for ssl authtype to log into Koji.
    # In Koji configuration, kerberos is the default authtype. If this is set,
    # ssl authtype will be used instead.
    # A workable value could be '/etc/mts/msg-tagger.pem'
    koji_cert = None

    # Used for kerberos authtype to log into Koji.
    # Example: '/etc/mts/mts.keytab'
    keytab = None
    # MTS host principal inside the keytab. If keytab is specified to use a
    # keytab explicitly, principal must be set as well.
    # Example: 'mts/hostname@EXAMPLE.COM'
    principal = None

    # Please note that, if neither keytab nor principal is set or valid, the
    # default or configured Kerberos ccache will be used to get ticket. That
    # means kinit should be run with the keytab and principal in advance.

    # Messages sent to these topics will be handled.
    # For internal, it is set to a UMB queue name, for example,
    # Consumer.client-mts.queue.VirtualTopic.eng.mbs.module.state.change
    consumer_topics = [
        'org.fedoraproject.prod.mbs.module.state.change',
    ]

    # Indicate which messaging backend will be used to send message.
    # Choices: fedmsg and rhmsg. For internal, set it to rhmsg.
    messaging_backend = 'fedmsg'

    # Define messaging backends for sending messages to message bus. In Fedora,
    # it is fedmsg, rhmsg for internal.
    messaging_backends = {
        'fedmsg': {
            'service': 'mts',
        },
        'rhmsg': {
            # Broker URIs to connect, e.g. ['amqps://host:5671', 'amqps://anotherhost:5671']
            'brokers': [],
            # Absolute path to certificate file used to authenticate freshmaker
            'certificate': '',
            # Absolute path to private key file used to authenticate freshmaker
            'private_key': '',
            # Absolute path to trusted CA certificate bundle.
            'ca_cert': '',
            # Prefix to construct full topic to send message
            'topic_prefix': 'VirtualTopic.eng.mts',
        },
    }

    # Default is INFO. Refer to Python logging module to know valid values.
    log_level = 'INFO'

    # A URL of rules file which can be accessible via HTTP GET without authentication.
    # Example: https://example.com/rules/mts-rules.yaml
    rules_file_url = ''


class DevConfiguration(BaseConfiguration):
    koji_profile = 'stg'
    consumer_topics = [
        'org.fedoraproject.dev.mbs.module.state.change',
        'org.fedoraproject.stg.mbs.module.state.change',
    ]
    log_level = 'DEBUG'
    rules_file_url = (
        'https://raw.githubusercontent.com/fedora-modularity/message-tagging-service/'
        'master/rules/mts-rules.yaml'
    )


class TestConfiguration(DevConfiguration):
    pass



# -*- coding: utf-8 -*-
#
# Message tagging service is an event-driven service to tag build.
# Copyright (C) 2019  Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Chenxiong Qi <cqi@redhat.com>

import fedmsg.consumers
import logging
import requests

from message_tagging_service import conf
from message_tagging_service import tagging_service
from message_tagging_service.utils import read_rule_defs

logger = logging.getLogger(__name__)


class MTSConsumer(fedmsg.consumers.FedmsgConsumer):
    topic = conf.consumer_topics
    config_key = 'mts-consumer'

    def __init__(self, *args, **kwargs):
        super(MTSConsumer, self).__init__(*args, **kwargs)

    def consume(self, msg):
        logger.debug('Got message: %r', msg)

        event_msg = msg['body']['msg']
        if event_msg['state_name'] != 'ready':
            logger.info('Skip module build %s-%s-%s-%s as it is not in ready state yet.',
                        event_msg['name'], event_msg['stream'],
                        event_msg['version'], event_msg['context'])
            return

        try:
            rule_defs = read_rule_defs()
        except requests.exceptions.HTTPError:
            logger.exception('Failed to retrieve rules content.')
        else:
            # For an empty yaml file, YAML returns None. So, if the remote rule
            # file is empty, catch this case and skip to handle the tag.
            if rule_defs is not None:
                tagging_service.handle(rule_defs, event_msg)



# -*- coding: utf-8 -*-
#
# Message tagging service is an event-driven service to tag build.
# Copyright (C) 2019  Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Chenxiong Qi <cqi@redhat.com>

import requests
import yaml
import logging

from message_tagging_service import conf

logger = logging.getLogger(__name__)


def retrieve_modulemd_content(module_build_id):
    """Retrieve and return modulemd.txt from MBS

    :param int module_build_id: module build ID.
    :return: modulemd content.
    :rtype: str
    """
    api_url = conf.mbs_api_url.rstrip('/')
    resp = requests.get(f'{api_url}/module-builds/{module_build_id}', params={
        'verbose': True
    })
    resp.raise_for_status()
    return resp.json()['modulemd']


def read_rule_defs():
    """Read rule definiations from configured rule file

    :return: a rule file is a YAML file, which is read, parsed and returned as
        a mapping.
    :rtype: dict
    """
    r = requests.get(conf.rules_file_url)
    r.raise_for_status()
    return yaml.safe_load(r.text)



# -*- coding: utf-8 -*-

import koji
import os
import pytest

from mock import Mock
from mock import call
from mock import patch
from textwrap import dedent

from message_tagging_service import tagging_service
from message_tagging_service.utils import read_rule_defs

test_data_dir = os.path.join(os.path.dirname(__file__), 'data')


# Fake koji config returned from koji.read_config that ensures krb_login is
# called successfully.
koji_config_krb_auth = {
    'authtype': 'kerberos',
    'debug': False,
    'server': 'http://localhost/',
    'cert': '',
    'keytab': None,
    'principal': None,
}


class TestRuleDefinitionCheck(object):
    """Test rule_matches_module_build"""

    def test_simple_rule_definition(self):
        rule_def = {
            'id': 'Simple match any module',
            'type': 'module',
            'description': 'Match any module build.',
            'destinations': 'modular-fallback-tag'
        }
        match = tagging_service.RuleDef(rule_def).match(Mock())
        assert match
        assert ['modular-fallback-tag'] == match.dest_tags

    def test_match_scratch_module_build(self):
        rule_def = {
            'id': 'Simple match by scratch',
            'type': 'module',
            'rule': {'scratch': True},
            'description': 'Match module build by scratch.',
            'destinations': 'modular-fallback-tag'
        }

        modulemd = {'data': {'scratch': True}}
        match = tagging_service.RuleDef(rule_def).match(modulemd)
        assert match
        assert ['modular-fallback-tag'] == match.dest_tags

        modulemd = {'data': {'scratch': False}}
        assert not tagging_service.RuleDef(rule_def).match(modulemd)

        modulemd = {'data': {}}
        assert not tagging_service.RuleDef(rule_def).match(modulemd)

    def test_match_development_module_build(self):
        rule_def = {
            'id': 'Simple match by development',
            'type': 'module',
            'rule': {'development': True},
            'description': 'Match module build by development.',
            'destinations': 'modular-fallback-tag'
        }

        modulemd = {'data': {'development': True}}
        match = tagging_service.RuleDef(rule_def).match(modulemd)
        assert match
        assert ['modular-fallback-tag'] == match.dest_tags

        modulemd = {'data': {'development': False}}
        assert not tagging_service.RuleDef(rule_def).match(modulemd)

        modulemd = {'data': {}}
        assert not tagging_service.RuleDef(rule_def).match(modulemd)

    def test_match_module_by_list_of_regex(self):
        rule_def = {
            'id': 'Match module by list of regular expressions',
            'type': 'module',
            'rule': {
                'name': [r'^javapackages-tools$', r'-ursamajor$']
            },
            'description': 'Match module build by development.',
            'destinations': r'\g<platform>-modular-ursamajor',
        }

        modulemd = {'data': {'name': 'javapackages-tools'}}
        match = tagging_service.RuleDef(rule_def).match(modulemd)
        assert match
        assert [r'\g<platform>-modular-ursamajor'] == match.dest_tags

        modulemd = {'data': {'name': 'module-a-ursamajor'}}
        assert tagging_service.RuleDef(rule_def).match(modulemd)

        modulemd = {'data': {'name': 'module-b'}}
        assert not tagging_service.RuleDef(rule_def).match(modulemd)

    def test_match_module_by_dict_type_rule(self):
        rule_def = {
            'id': 'Match module by name',
            'type': 'module',
            'rule': {
                'dependencies': {
                    'buildrequires': {'platform': r'^.*$'},
                    'requires': {'platform': r'^(?P<platform>f\d+)$'}
                },
            },
            'description': 'Match module build by development.',
            'destinations': r'\g<platform>-modular-ursamajor',
        }

        modulemd = {'data': {
            'dependencies': [{
                'buildrequires': {'platform': ['f28']},
                'requires': {'platform': ['f28']},
            }]
        }}
        match = tagging_service.RuleDef(rule_def).match(modulemd)
        assert match
        assert ['f28-modular-ursamajor'] == match.dest_tags


class TestMatchRuleDefinitions(object):

    @patch('message_tagging_service.tagging_service.retrieve_modulemd_content')
    @patch('message_tagging_service.tagging_service.tag_build')
    def test_not_tag_build_if_no_rules_are_matched(
            self, tag_build, retrieve_modulemd_content):
        # Note that, platform does not match the rule in rule file.
        retrieve_modulemd_content.return_value = dedent('''\
            ---
            document: modulemd
            version: 2
            data:
              name: ant
              stream: 1
              version: 1
              context: c1
              dependencies:
              - buildrequires:
                  platform: [el8]
                requires:
                  platform: [el8]
            ''')

        rule_file = os.path.join(test_data_dir, 'mts-test-for-no-match.yaml')
        with patch('requests.get') as get:
            with open(rule_file, 'r') as f:
                get.return_value.text = f.read()
            rule_defs = read_rule_defs()

        with patch.object(tagging_service.logger, 'info') as info:
            # Note that, no module property matches rule in rule file.
            tagging_service.handle(rule_defs, {
                'id': 1,
                'name': 'ant',
                'stream': '1',
                'version': '1',
                'context': 'c1',
                'state_name': 'ready',
            })

            assert ('Module build %s does not match any rule.', 'ant-1-1-c1') == \
                info.call_args[0]
            tag_build.assert_not_called()

    @patch('message_tagging_service.tagging_service.retrieve_modulemd_content')
    @patch('message_tagging_service.messaging.publish')
    @patch('koji.ClientSession')
    @patch('koji.read_config')
    def test_tag_build_if_match_one_rule_only(
            self, read_config, ClientSession, publish, retrieve_modulemd_content):
        read_config.return_value = koji_config_krb_auth

        # Note that, platform does not match the rule in rule file.
        retrieve_modulemd_content.return_value = dedent('''\
            ---
            document: modulemd
            version: 2
            data:
              name: javapackages-tools
              stream: 1
              version: 1
              context: c1
              dependencies:
              - buildrequires:
                  platform: [f29]
                requires:
                  platform: [f29]
            ''')

        rule_file = os.path.join(test_data_dir, 'mts-test-rules.yaml')
        with patch('requests.get') as get:
            with open(rule_file, 'r') as f:
                get.return_value.text = f.read()
            rule_defs = read_rule_defs()

            session = ClientSession.return_value
            session.tagBuild.side_effect = [1, 2, 3]

            tagging_service.handle(rule_defs, {
                'id': 1,
                'name': 'javapackages-tools',
                'stream': '1',
                'version': '1',
                'context': 'c1',
                'state_name': 'ready',
            })

            nvr = 'javapackages-tools-1-1.c1'
            nvr_devel = 'javapackages-tools-devel-1-1.c1'
            session.tagBuild.assert_has_calls([
                call('f29-modular-ursamajor', nvr),
                call('f29-modular-ursamajor', nvr_devel),
            ], any_order=True)

            # 2 messages should be sent:
            # javapackages-tools: f29
            # javapackages-tools-devel: f29
            publish.assert_has_calls([
                call('build.tag.requested', {
                    'build': {
                        'id': 1, 'name': 'javapackages-tools',
                        'stream': '1', 'version': '1', 'context': 'c1',
                    },
                    'nvr': nvr,
                    'destination_tags': [
                        {'tag': 'f29-modular-ursamajor', 'task_id': 1},
                    ],
                }),
                call('build.tag.requested', {
                    'build': {
                        'id': 1, 'name': 'javapackages-tools-devel',
                        'stream': '1', 'version': '1', 'context': 'c1',
                    },
                    'nvr': nvr_devel,
                    'destination_tags': [
                        {'tag': 'f29-modular-ursamajor', 'task_id': 2},
                    ],
                }),
            ], any_order=True)

    @pytest.mark.parametrize(
        'tagBuild_side_effect,expected_destination_tags_for_build,'
        'expected_destination_tags_for_devel_build',
        [
            # All tagBuild requests succeed
            [
                [1, 2, 3, 4, 5],
                [
                    {'tag': 'f29-modular-ursamajor', 'task_id': 1},
                    {'tag': 'f28-modular-ursamajor', 'task_id': 2},
                ],
                [
                    {'tag': 'f29-modular-ursamajor', 'task_id': 3},
                    {'tag': 'f28-modular-ursamajor', 'task_id': 4},
                ]
            ],
            # All tagBuild requests fail
            [
                [
                    koji.TagError('failed to tag build'),
                    koji.TagError('failed to tag build'),
                    koji.TagError('failed to tag build'),
                    koji.TagError('failed to tag build'),
                ],
                [
                    {
                        'tag': 'f29-modular-ursamajor',
                        'task_id': None,
                        'error': 'failed to tag build'
                    },
                    {
                        'tag': 'f28-modular-ursamajor',
                        'task_id': None,
                        'error': 'failed to tag build'
                    },
                ],
                [
                    {
                        'tag': 'f29-modular-ursamajor',
                        'task_id': None,
                        'error': 'failed to tag build'
                    },
                    {
                        'tag': 'f28-modular-ursamajor',
                        'task_id': None,
                        'error': 'failed to tag build'
                    },
                ]
            ],
            # tagBuild requests fail for f29-* tags, but succeed for f28-*
            [
                [
                    koji.TagError('failed to tag build'),
                    1,  # Returned fake task id for applying tag f28-* to build
                    koji.TagError('failed to tag build'),
                    2,  # Returned fake task id for applying tag f28-* to devel build
                ],
                [
                    {
                        'tag': 'f29-modular-ursamajor',
                        'task_id': None,
                        'error': 'failed to tag build'
                    },
                    {
                        'tag': 'f28-modular-ursamajor',
                        'task_id': 1,
                    },
                ],
                [
                    {
                        'tag': 'f29-modular-ursamajor',
                        'task_id': None,
                        'error': 'failed to tag build'
                    },
                    {
                        'tag': 'f28-modular-ursamajor',
                        'task_id': 2,
                    },
                ]
            ],
        ])
    @patch('message_tagging_service.tagging_service.retrieve_modulemd_content')
    @patch('message_tagging_service.messaging.publish')
    @patch('koji.ClientSession')
    @patch('koji.read_config')
    def test_apply_multiple_tags_to_a_matched_build(
            self, read_config, ClientSession, publish, retrieve_modulemd_content,
            tagBuild_side_effect, expected_destination_tags_for_build,
            expected_destination_tags_for_devel_build
    ):
        read_config.return_value = koji_config_krb_auth

        # Because rule file specifies that destination tag uses the value of
        # requires.platform, and there are two of those values in modulemd,
        # the module build should be tagged with two tags.
        retrieve_modulemd_content.return_value = dedent('''\
            ---
            document: modulemd
            version: 2
            data:
              name: javapackages-tools
              stream: 1
              version: 1
              context: c1
              dependencies:
              - buildrequires:
                  platform: [f29]
                requires:
                  platform: [f29, f28]
            ''')

        session = ClientSession.return_value
        session.tagBuild.side_effect = tagBuild_side_effect

        rule_file = os.path.join(test_data_dir, 'mts-test-rules.yaml')
        with patch('requests.get') as get:
            with open(rule_file, 'r') as f:
                get.return_value.text = f.read()

            rule_defs = read_rule_defs()

            tagging_service.handle(rule_defs, {
                'id': 1,
                'name': 'javapackages-tools',
                'stream': '1',
                'version': '1',
                'context': 'c1',
                'state_name': 'ready',
            })

            nvr = 'javapackages-tools-1-1.c1'
            nvr_devel = 'javapackages-tools-devel-1-1.c1'
            session.tagBuild.assert_has_calls([
                call('f29-modular-ursamajor', nvr),
                call('f28-modular-ursamajor', nvr),
                call('f29-modular-ursamajor', nvr_devel),
                call('f28-modular-ursamajor', nvr_devel),
            ], any_order=True)

            # 2 messages should be sent:
            # javapackages-tools: f29, f28
            # javapackages-tools-devel: f29, f28
            publish.assert_has_calls([
                call('build.tag.requested', {
                    'build': {
                        'id': 1, 'name': 'javapackages-tools',
                        'stream': '1', 'version': '1', 'context': 'c1',
                    },
                    'nvr': nvr,
                    'destination_tags': expected_destination_tags_for_build,
                }),
                call('build.tag.requested', {
                    'build': {
                        'id': 1, 'name': 'javapackages-tools-devel',
                        'stream': '1', 'version': '1', 'context': 'c1',
                    },
                    'nvr': nvr_devel,
                    'destination_tags': expected_destination_tags_for_devel_build,
                }),
            ], any_order=True)


class TestLoginKoji(object):
    """Test login_koji"""

    @patch.object(tagging_service.conf, 'koji_cert', new='path/to/cert')
    @patch('os.path.exists', return_value=True)
    @patch('os.access', return_value=True)
    def test_ssl_login(self, access, exists):
        session = Mock()
        # Ensure koji_cli.lib.activate_session completes API version check.
        session.getAPIVersion.return_value = 1

        tagging_service.login_koji(session, {
            'authtype': 'kerberos',
            'serverca': '',
            'debug': False,
        })

        session.ssl_login.assert_called_once_with(
            'path/to/cert', None, '', proxyuser=None)

    @pytest.mark.parametrize("keytab,principal,krb_login_kwargs", [
        (None, None, {'proxyuser': None}),
        ('path/to/keytab', None, {'proxyuser': None}),
        (None, 'mts/hostname@EXAMPLE.COM', {'proxyuser': None}),
        ('path/to/keytab', 'mts/hostname@EXAMPLE.COM', {
            'keytab': 'path/to/keytab',
            'principal': 'mts/hostname@EXAMPLE.COM',
            'proxyuser': None
        }),
    ])
    @patch('os.path.exists', return_value=True)
    @patch('os.access', return_value=True)
    def test_krb_login(self, access, exists, keytab, principal, krb_login_kwargs):
        session = Mock()
        session.getAPIVersion.return_value = 1

        with patch.object(tagging_service.conf, 'keytab', new=keytab):
            with patch.object(tagging_service.conf, 'principal', new=principal):
                tagging_service.login_koji(session, koji_config_krb_auth)

        session.krb_login.assert_called_once_with(**krb_login_kwargs)

    @patch.object(tagging_service.conf, 'koji_cert', new='path/to/cert')
    def test_raise_error_if_ssl_cert_is_not_readable(self):
        session = Mock()
        # Ensure koji_cli.lib.activate_session completes API version check.
        session.getAPIVersion.return_value = 1

        with pytest.raises(
                IOError, message='SSL certificate path/to/cert is not readable.'):
            tagging_service.login_koji(session, {
                'authtype': 'kerberos',
                'serverca': '',
                'debug': False,
            })



# -*- coding: utf-8 -*-
#
# Message tagging service is an event-driven service to tag build.
# Copyright (C) 2019  Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Troy Dawson
#          Chenxiong Qi <cqi@redhat.com>

import koji
import koji_cli.lib
import logging
import os
import re
import requests
import yaml

from collections import namedtuple
from contextlib import contextmanager

from message_tagging_service import conf
from message_tagging_service import messaging
from message_tagging_service import monitor
from message_tagging_service.utils import retrieve_modulemd_content

logger = logging.getLogger(__name__)

TagBuildResult = namedtuple('TagBuildResult', ['tag_name', 'task_id', 'error'])


class RuleMatch(object):
    """Result of :meth:`RuleDef.match`

    A rule match object can be evaluated as truth or false. Hence, it is doable
    like this::

        match = RuleDef.match(...)
        if match: ...
        if not match: ...

    :param bool matched: indicate if rule definition is matched.
    :param str dest_tags: the formatted destination tags from rule definition
        item ``destination``.
    """

    def __init__(self, matched, dest_tags=None):
        self.matched = matched
        self.dest_tags = dest_tags

    def __bool__(self):
        return self.matched


class RuleDef(object):
    """Represent a rule definition

    Refer to https://pagure.io/modularity/blob/master/f/drafts/module-tagging-service/format.md
    """

    def __init__(self, data):
        for name in ['id', 'type', 'destinations']:
            if name not in data:
                raise ValueError(f'Rule definition does not have property {name}.')

        self.data = data

        self._property_matches = []
        self._regex_has_named_group = []

    @property
    def id(self):
        return self.data['id']

    @property
    def description(self):
        return self.data['description']

    @property
    def type(self):
        return self.data['type']

    @property
    def rule(self):
        """Return property rule of definition

        Note that, a rule definition may or may not have match criteria in rule
        property. If no rule is defined, None will be returned.
        """
        # YAML allows to read a empty section like:
        # - name: xxx
        #   rule:
        #   destination: xxx
        # In this case, parsed YAML dict has key/value: {'rule': None}
        return self.data.get('rule')

    @property
    def destinations(self):
        return self.data['destinations']

    def find_diff_value(self, regex, mmd_property_value):
        """Match a property value with expected regular expression

        :param str regex: the regular expression to try to match property value.
        :param mmd_property_value: property value to match. It could be a single
            value, or a list of values for example the dependencies like
            ``{'dependencies': {'buildrequires': {'platform': ['f28']}}}``.
        :type mmd_property_value: list or str
        :return: True if given regular expression matches the single value, or
            match one of the list of values. If not match anything, False is
            returned.
        :rtype: bool
        """
        if isinstance(mmd_property_value, list):
            check_values = mmd_property_value
        else:
            check_values = [mmd_property_value]

        matches_found = False
        for value in check_values:
            match = re.search(regex, value)
            if match:
                matches_found = True
                if match.groupdict():
                    self._regex_has_named_group.append((regex, value))

        return matches_found

    def find_diff_list(self, match_candidates, mmd_property_value):
        """Find out if module property value matches one of values defined in rule

        :param match_candidates: list of regular expressions in rule definition
            to check if one of them could match the corresponding property
            value in modulemd.
        :type match_candidates: list[str]
        :param str mmd_property_value: modulemd's property value to check.
        :return: True if match, otherwise False.
        :rtype: bool
        """
        logger.debug('Checking %s against regular expressions %r',
                     mmd_property_value, match_candidates)
        for regex in match_candidates:
            if self.find_diff_value(regex, mmd_property_value):
                return True
        return False

    def find_diff_dict(self, rule_dict, check_dict):
        """Check if rule matches modulemd values recursively for dict type

        Modulemd dependencies is a dict, which could be::

            {
                'dependencies': {
                    'buildrequires': {'platform': ['f29']},
                    'requires': {'platform': ['f29']}
                }
            }

        When rule definition has a rule like::

            {'dependencies': {'requires': {'platform': r'f\d+'}}}

        this function has to check if rule matches the module's runtime
        requirement platform f29.

        :param dict rule_dict:
        :param dict check_dict:
        :return: True if match, otherwise False.
        :rtype: bool
        """  # noqa
        match = True
        for key, value in rule_dict.items():
            new_check_dict = check_dict.get(key)
            if new_check_dict is None:
                logger.debug("'%s' is not found in module", key)
                return False
            if isinstance(value, dict):
                match = self.find_diff_dict(value, new_check_dict)
            elif isinstance(value, list):
                match = self.find_diff_list(value, new_check_dict)
            else:
                match = self.find_diff_value(value, new_check_dict)
            if not match:
                # As long as one of rule criteria does not match module
                # property, the whole dict rule match fails.
                break
        return match

    def match(self, modulemd):
        """Check if a rule definition matches a module

        The match implementation follows
        https://pagure.io/modularity/blob/master/f/drafts/module-tagging-service/format.md

        :param dict modulemd: a mapping parsed from modulemd YAML file.
        :return: a RuleMatch object to indicate whether modulemd matches the rule.
        :rtype: :class:`RuleMatch`
        """
        if self.rule is None:
            logger.debug(
                'No rule criteria is defined. Build will be tagged to %s',
                self.destinations)
            return RuleMatch(True, [self.destinations])

        for property, expected in self.rule.items():
            # Both scratch and development have default value to compare with
            # expected in rule definition.

            if property == 'scratch':
                mmd_value = modulemd['data'].get("scratch", False)
                if expected == mmd_value:
                    logger.debug('Rule/Value: %s: %s. Matched.', property, expected)
                    self._property_matches.append(True)
                else:
                    logger.debug('Rule/Value: %s: %s. Not Matched. Real value: %s',
                                 property, expected, mmd_value)
                    self._property_matches.append(False)

            elif property == 'development':
                mmd_value = modulemd["data"].get("development", False)
                if expected == mmd_value:
                    logger.debug('Rule/Value: %s: %s. Matched.', property, expected)
                    self._property_matches.append(True)
                else:
                    logger.debug('Rule/Value: %s: %s. Not Matched. Real value: %s',
                                 property, expected, mmd_value)
                    self._property_matches.append(False)

            else:
                # Now check rules that have regex
                value_to_check = modulemd["data"].get(property)
                if value_to_check is None:
                    logger.debug('%s is not match. Modulemd does not have %s', property, property)
                    self._property_matches.append(False)

                elif isinstance(expected, dict):
                    if self.find_diff_dict(expected, value_to_check[0]):
                        logger.debug('Rule/Value: %s: %r. Matched.', property, expected)
                        self._property_matches.append(True)
                    else:
                        logger.debug('Rule/Value: %s: %r. Not Matched. Real value: %r',
                                     property, expected, value_to_check[0])
                        self._property_matches.append(False)

                elif isinstance(expected, list):
                    if self.find_diff_list(expected, value_to_check):
                        logger.debug('Rule/Value: %s: %r. Matched.', property, expected)
                        self._property_matches.append(True)
                    else:
                        logger.debug('Rule/Value: %s: %r. Not Matched. Real value: %s',
                                     property, expected, value_to_check)
                        self._property_matches.append(False)

                else:
                    if self.find_diff_value(expected, str(value_to_check)):
                        logger.debug('Rule/Value: %s: %r. Matched.', property, expected)
                        self._property_matches.append(True)
                    else:
                        logger.debug('Rule/Value: %s: %r. Not Matched.', property, expected)
                        self._property_matches.append(False)

        if all(self._property_matches):
            if self._regex_has_named_group:
                formatted_dest_tags = [
                    re.sub(regex, self.destinations, mmd_property_value)
                    for regex, mmd_property_value in self._regex_has_named_group
                ]
                return RuleMatch(True, formatted_dest_tags)
            else:
                return RuleMatch(True, [self.destinations])
        else:
            return RuleMatch(False)


def login_koji(session, config):
    """Log into Koji

    By default, client is authenticated via Kerberos, which is set in Koji
    configuration. However, it could be changed to autheticate client via SSL
    by setting ``koji_cert`` in MTS configuration.

    Please note that, only both ``keytab`` and ``principal`` are set in
    configuration, the session could be logged in with a specific keytab.

    :param session: an object of ClientSession to login.
    :type session: koji.ClientSession
    :param dict config: a mapping containing koji config. Generally, it is the
        return value from ``koji.read_config``.
    """
    # Loaded koji config is a simple dict without nested dict. dict.copy just
    # works well.
    cfg = config.copy()
    use_ssl = conf.koji_cert is not None
    if use_ssl:
        if os.path.exists(conf.koji_cert) and os.access(conf.koji_cert, os.R_OK):
            logger.info('conf.koji_cert is set. Use ssl authtype.')
            cfg['cert'] = conf.koji_cert
            cfg['authtype'] = 'ssl'
        else:
            raise IOError(f'SSL certificate {conf.koji_cert} is not readable.')
    else:
        # otherwise, do the default authtype: kerberos
        if (conf.keytab and os.path.exists(conf.keytab) and
                os.access(conf.keytab, os.R_OK) and conf.principal):
            cfg['keytab'] = conf.keytab
            cfg['principal'] = conf.principal

    return koji_cli.lib.activate_session(session, cfg)


@contextmanager
def make_koji_session():
    koji_config = koji.read_config(conf.koji_profile)
    session_opts = koji.grab_session_options(koji_config)
    koji_session = koji.ClientSession(koji_config['server'], opts=session_opts)
    login_koji(koji_session, koji_config)

    yield koji_session

    koji_session.logout()


def tag_build(nvr, dest_tags, koji_session):
    """Tag build with specific tags

    Calling Koji API to tag build might fail, however successful tagged tag will
    be returned and to log the failed tag operation.

    :param str nvr: build NVR.
    :param dest_tags: tag names.
    :type dest_tags: list[str]
    :return: a list of tag build result info, each of them is an object of
        ``TagBuildResult``. The first element is tag name to apply, the second
        one is the task id return from Koji, and the last one is the error
        message. If tag operation is requested successfully, error message is
        set to None, otherwise None is set to task id and error message has
        some content.
    :rtype: list[TagBuildResult]
    """
    tagged_tags = []
    for tag in dest_tags:
        try:
            if conf.dry_run:
                logger.info("DRY-RUN: koji_session.tagBuild('%s', '%s')", tag, nvr)
                task_id = 1
            else:
                task_id = koji_session.tagBuild(tag, nvr)
        except Exception as e:
            tagged_tags.append(
                TagBuildResult(tag_name=tag, task_id=None, error=str(e)))
            monitor.failed_tag_build_requests_counter.inc()
        else:
            tagged_tags.append(
                TagBuildResult(tag_name=tag, task_id=task_id, error=None))
    return tagged_tags


def handle(rule_defs, event_msg):
    this_name = event_msg["name"]
    this_stream = event_msg["stream"]
    this_version = event_msg["version"]
    this_context = event_msg["context"]
    nsvc = f"{this_name}-{this_stream}-{this_version}-{this_context}"

    try:
        modulemd = yaml.safe_load(retrieve_modulemd_content(event_msg['id']))
    except requests.exceptions.HTTPError as e:
        logger.exception(f'Failed to retrieve modulemd for {nsvc}: {str(e)}')

        # Continue to wait for and handle next module build which moves
        # to ready state.
        return

    logger.debug('Modulemd file is downloaded and parsed.')

    rule_match = None
    for i, rule_def in enumerate(rule_defs, 1):
        rd = RuleDef(rule_def)
        logger.info('[%s] Checking rule definition: %s', i, rd.id)
        match = rd.match(modulemd)
        if match:
            logger.info('[%d] Rule definition: Matched. Remaining rules ignored.', i)
            rule_match = match
            break
        else:
            logger.info('[%d] Rule definition: Not Matched.', i)

    if not rule_match:
        logger.info('Module build %s does not match any rule.', nsvc)
        messaging.publish('build.tag.unmatched', {
            'build': {
                'id': event_msg['id'],
                'name': this_name,
                'stream': this_stream,
                'version': this_version,
                'context': this_context,
            },
        })
        return

    monitor.matched_module_builds_counter.inc()

    stream = this_stream.replace('-', '_')
    with make_koji_session() as koji_session:
        for name in (this_name, f'{this_name}-devel'):
            nvr = f'{name}-{stream}-{this_version}.{this_context}'

            dest_tags = rule_match.dest_tags
            logger.info('Tag build %s with tag(s) %s', nvr, ', '.join(dest_tags))
            tag_build_result = tag_build(nvr, dest_tags, koji_session)

            failed_tasks = [item for item in tag_build_result if item.task_id is None]

            if len(failed_tasks) == len(dest_tags):
                logger.warning(
                    'None of tag(s) %r is applied to build %s successfully.',
                    dest_tags, nvr)
            elif len(failed_tasks) > 0:
                logger.warning(
                    'Tag(s) %r should be applied to build %s. But some of them,'
                    ' %s, failed to be applied.',
                    nvr, dest_tags, [item[0] for item in failed_tasks])

            # Tag info for message sent later
            # For a successful tag task, it is {"tag": "name", "task_id": 123}
            # For a failure tag task, it is {"tag": "name", "task_id": None, "reason": "..."}
            destination_tags = []
            for result in tag_build_result:
                data = {'tag': result.tag_name, 'task_id': result.task_id}
                if result.task_id is None:
                    data['error'] = result.error
                destination_tags.append(data)

            messaging.publish('build.tag.requested', {
                'build': {
                    'id': event_msg['id'],
                    'name': name,
                    'stream': this_stream,
                    'version': this_version,
                    'context': this_context,
                },
                'nvr': nvr,
                'destination_tags': destination_tags,
            })



# -*- coding: utf-8 -*-

from setuptools import find_packages, setup


with open('README.rst', 'r') as f:
    long_description = f.read()


with open('requirements.txt', 'r') as f:
    install_requires = [
        line.rstrip() for line in f if not line.startswith('#')
    ]


setup(
    name='message-tagging-service',
    version='0.4',
    description='Tag Koji build with correct tag which is triggered by message bus',
    long_description=long_description,
    maintainer='Factory 2 Team',
    # Not sure which email address is proper here, so just using a fake email
    # address. But, probably this could be a real some day, or another one.
    maintainer_email='message-tagging-service@fedoraproject.org',
    license='GPLv2+',
    url='https://github.com/fedora-modularity/message-tagging-service',
    keywords='messaging,module,tag,koji',
    packages=find_packages(),
    install_requires=install_requires,
    entry_points='''
        [moksha.consumer]
        mts-consumer = message_tagging_service.consumer:MTSConsumer
    ''',
    classifiers=[
        'License :: OSI Approved :: GNU Lesser General Public License v2 or later (LGPLv2+)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    data_files=[
        ('/etc/mts/', ['conf/config.py']),
    ],
)



# -*- coding: utf-8 -*-
#
# Message tagging service is an event-driven service to tag build.
# Copyright (C) 2019  Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Chenxiong Qi <cqi@redhat.com>

import json
import pytest

from message_tagging_service import messaging
from mock import patch

try:
    import rhmsg
except ImportError:
    rhmsg = None


class TestMessaging(object):

    @patch.object(messaging.conf, 'messaging_backend', new='fedmsg')
    @patch.dict(messaging.conf.messaging_backends, values={'fedmsg': {
        'service': 'myapp'
    }})
    @patch('fedmsg.publish')
    def test_send_via_fedmsg(self, publish):
        messaging.publish('build.tagged', {'build_id': 1})

        publish.assert_called_once_with(
            'build.tagged', msg={'build_id': 1}, modname='myapp')

    @pytest.mark.skipif(not rhmsg, reason='Library rhmsg is not available.')
    @patch.object(messaging.conf, 'messaging_backend', new='rhmsg')
    @patch.dict(messaging.conf.messaging_backends, values={'rhmsg': {
        'topic_prefix': 'VirtualTopic.eng.mts.',
        'brokers': ['amqps://broker1/', 'amqps://broker2/'],
        'certificate': '/path/to/certificate',
        'private_key': '/path/to/private_key',
        'ca_cert': '/path/to/ca_cert',
    }})
    @patch('proton.Message')
    @patch('rhmsg.activemq.producer.AMQProducer')
    def test_send_via_rhmsg(self, AMQProducer, Message):
        msg = {'koji_tag': 'module-a-1-1-c1'}
        messaging.publish('build.tagged', msg)

        config = {
            'urls': ['amqps://broker1/', 'amqps://broker2/'],
            'certificate': '/path/to/certificate',
            'private_key': '/path/to/private_key',
            'trusted_certificates': '/path/to/ca_cert',
        }
        AMQProducer.assert_called_once_with(**config)
        producer = AMQProducer.return_value.__enter__.return_value
        producer.through_topic.assert_called_once_with('VirtualTopic.eng.mts.build.tagged')

        Message.return_value.body = json.dumps(msg)
        producer.send.assert_called_once_with(Message.return_value)

    @patch.object(messaging.conf, 'messaging_backend', new='anothercool')
    def test_no_backend_handler_is_found(self):
        with pytest.raises(KeyError):
            messaging.publish('topic', {})



# -*- coding: utf-8 -*-
#
# Message tagging service is an event-driven service to tag build.
# Copyright (C) 2019  Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Chenxiong Qi

import queue

from mock import MagicMock
from mock import patch

from message_tagging_service import consumer


class TestConsumer(object):

    def new_consumer(self):
        hub = MagicMock()
        hub.config = {}
        hub.config['mts-consumer'] = True
        hub.config['validate_signatures'] = False
        c = consumer.MTSConsumer(hub)
        c.incoming = queue.Queue()
        return c

    @patch.object(consumer, 'read_rule_defs')
    @patch.object(consumer.tagging_service, 'handle')
    def test_consume_msg(self, handle, read_rule_defs):
        consumer = self.new_consumer()

        event_msg = {'state_name': 'ready'}
        consumer.consume({'body': {'msg': event_msg}})

        handle.assert_called_once_with(read_rule_defs.return_value, event_msg)

    @patch.object(consumer, 'read_rule_defs')
    @patch.object(consumer.tagging_service, 'handle')
    def test_ignore_message_if_not_ready(self, handle, read_rule_defs):
        consumer = self.new_consumer()
        consumer.consume({'body': {'msg': {
            'name': 'modulea',
            'stream': '10',
            'version': '201902141503',
            'context': '00000000',
            'state_name': 'init',
            'koji_tag': None,
        }}})

        handle.assert_not_called()

    @patch('requests.get')
    @patch.object(consumer.tagging_service, 'handle')
    def test_skip_if_rule_file_is_empty(self, handle, get):
        get.return_value.text = '---'

        consumer = self.new_consumer()
        consumer.consume({'body': {'msg': {
            'state_name': 'ready',
            'koji_tag': 'module-modulea-1-1-c1',
        }}})

        handle.assert_not_called()



# -*- coding: utf-8 -*-
#
# Message tagging service is an event-driven service to tag build.
# Copyright (C) 2019  Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Original authors: Filip Valder

# Copied from https://pagure.io/monitor-flask-sqlalchemy/blob/master/f/monitor.py
# Customized for MTS, e.g. database metrics are removed because MTS does not
# uses a database.

import os
import tempfile

from prometheus_client import CollectorRegistry
from prometheus_client import ProcessCollector
from prometheus_client import multiprocess
from prometheus_client import Counter
from prometheus_client import generate_latest

if not os.environ.get('prometheus_multiproc_dir'):
    dir_name = tempfile.mkdtemp(prefix='mts-prometheus-multiproc-')
    os.environ.setdefault('prometheus_multiproc_dir', dir_name)

registry = CollectorRegistry()
ProcessCollector(registry=registry)
multiprocess.MultiProcessCollector(registry)

failed_tag_build_requests_counter = Counter(
    'failed_tag_build_requests',
    'The number of failed tagBuild API calls.',
    registry=registry
)

matched_module_builds_counter = Counter(
    'matched_module_builds',
    'The number of module builds which are matched rule(s) to be tagged.',
    registry=registry
)

messages_notify_errors_counter = Counter(
    'messages_notify_errors',
    'The number of errors occurred during sending message to bus.',
    registry=registry
)


def generate_metrics_report():
    return generate_latest(registry)
